/*
   Windows Switcher to iGPU for Retina Macbook Pro systems
   Proof of Concept Code

   WARNING: This driver code is only for testing purposes, not for production use
*/

#include "gmuxtest.h"

#define NT_DEVICE_NAME L"\\Driver\\GMUXTEST"
#define DOS_DEVICE_NAME L"\\DosDevices\\GMUXTEST"

typedef struct _DEV_EXT {
  PVOID CbRegistration;
  PCALLBACK_OBJECT CbObject;
} DEV_EXT, *PDEV_EXT;

DEV_EXT gDevExt;

DRIVER_INITIALIZE DriverEntry;

__drv_dispatchType(IRP_MJ_CREATE)
__drv_dispatchType(IRP_MJ_CLOSE)
DRIVER_DISPATCH dummyFunction;
__drv_dispatchType(IRP_MJ_DEVICE_CONTROL)
DRIVER_DISPATCH deviceControl;
DRIVER_UNLOAD GMUXUnload;

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT,DriverEntry)
#pragma alloc_text(PAGE,GMUXUnload)
#pragma alloc_text(PAGE,dummyFunction)
#pragma alloc_text(PAGE,deviceControl)
#endif

#define GMUX_HARDCODED_IO_START 0x0700

#define GMUX_PORT_DPM_RADDR 0xd0
#define GMUX_PORT_DPM_WADDRSTAT 0xd4
#define GMUX_PORT_DPM_DAT0 0xc2
#define GMUX_PORT_DPM_DAT1 0xc3
#define GMUX_PORT_DPM_DAT2 0xc4
#define GMUX_PORT_DPM_DAT3 0xc5


#define GMUX_INTERRUPT_ENABLE		0xff

#define GMUX_PORT_INTERRUPT_ENABLE	0x14
#define GMUX_PORT_INTERRUPT_STATUS	0x16

#define GMUX_PORT_SWITCH_DISPLAY 0x10
#define GMUX_PORT_SWITCH_DDC 0x28
#define GMUX_PORT_SWITCH_EXTERNAL 0x40
#define GMUX_PORT_SWITCH_GET_EXTERNAL 0x41

#define GMUX_PORT_DISCRETE_POWER 0x50

#define GMUX_DPM_CMD_TO 20


/**************************************************************************
  IsIntelGPUEnabled
**************************************************************************/

UCHAR IsIntelGpuEnabled()
{
  //
  // This function checks if the Iris Pro graphics is enabled on this
  // system.  Under normal boot conditions, Haswell Retina Macbook Pro
  // EFI will hide Iris Pro GPU upon completion of the EFI BootServices
  // unless OS bootloader calls Apple-specific UEFI call which identifies
  // the OS as Mac OS X
  // 
  // We use this to detect whether user wants to use Intel Iris Pro
  // in Windows, as in this case iGPU will be enabled by GRUB apple_set_os
  // call

  UCHAR buf[8];
  ULONG offset = 0x54;
  ULONG length = 1;

  PCI_SLOT_NUMBER slot;
  ULONG busNumber;
  int error;

  busNumber = 0;
  
  slot.u.AsULONG = 0;
  slot.u.bits.DeviceNumber = 0;
  slot.u.bits.FunctionNumber = 0;
  
  error = HalGetBusDataByOffset(
    PCIConfiguration, 
    busNumber, 
    slot.u.AsULONG,
    buf, 
    offset, 
    length);


  if(error == 0)
  {
    //
    // PCI bus does not exist
    
    return 0;
  }
  else if(length != 2 && error == 2)
  {
    //
    // No device
    
    return 0;
  }
  else if(length != error)
  {
    //
    // Error reading config
    
    return 0;
  }

  //
  // Value of 0x0D means the GPU is disabled
  // Value of 0x3D means GPU is ebabled
  
  return (buf[0] == 0x3D) ? 1 : 0;
}

/**************************************************************************
  udelay
**************************************************************************/

void udelay(const LONGLONG microsec)
{
  LARGE_INTEGER li;
  li.QuadPart = microsec * 10; // 100-ns units

  KeDelayExecutionThread(
    KernelMode,
    FALSE,
    &li
    );
}

/**************************************************************************
  Port I/O
**************************************************************************/

UCHAR inb(USHORT port)
{
  PUCHAR paddr = (PUCHAR)port;
  return READ_PORT_UCHAR(paddr);
}

void outb(UCHAR val, USHORT port)
{
  PUCHAR paddr = (PUCHAR)port;
  WRITE_PORT_UCHAR(paddr, val);
}

/**************************************************************************
  GmuxDpmCmdReady
**************************************************************************/

UCHAR GmuxDpmCmdReady(void)
{
  UCHAR status = inb(GMUX_PORT_DPM_WADDRSTAT);
  UCHAR to = GMUX_DPM_CMD_TO;

  while (status == 0 && to)
  {
    inb(GMUX_PORT_DPM_RADDR);
    udelay(100000); // 100 ms
    status = inb(GMUX_PORT_DPM_WADDRSTAT);
    to--;
    udelay(100000); // 100 ms
  }
  return to != 0;
}

/**************************************************************************
  GmuxDpmCmdDone
**************************************************************************/

UCHAR GmuxDpmCmdDone(void)
{
  UCHAR status = 0;
  UCHAR to = GMUX_DPM_CMD_TO;

  do {
    status = inb(GMUX_PORT_DPM_WADDRSTAT);
    to--;
    udelay(100000); // 100 ms
  } while (status == 0 && to);

  if (status == 0)
    inb(GMUX_PORT_DPM_RADDR);

  return to != 0;
}

/**************************************************************************
  GmuxDpmWrite
**************************************************************************/

int GmuxDpmWrite(UCHAR reg, UCHAR size, ULONG val)
{
  if (size == 1)
  {
    outb((UCHAR)val, GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT0);
  }
  else if (size == 2)
  {
    outb((UCHAR)(val>>8), GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT1);
    outb((UCHAR)val, GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT0);
  }
  else if (size == 4)
  {
    outb((UCHAR)(val>>24), GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT3);
    outb((UCHAR)(val>>16), GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT2);
    outb((UCHAR)(val>>8), GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT1);
    outb((UCHAR)val, GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_DAT0);
  }
  else
  {
    return 0;
  }

  if (!GmuxDpmCmdReady())
  {
    return 0;
  }

  outb(reg, GMUX_HARDCODED_IO_START + GMUX_PORT_DPM_WADDRSTAT);

  if (!GmuxDpmCmdDone())
  {
    return 0;
  }

  return 1;
}

void GmuxWrite8(UCHAR port, UCHAR val)
{
  GmuxDpmWrite(port, 1, (ULONG)val);
}


VOID SwitchToIntelGraphicsAndPowerOffDiscreteGPU()
{
  GmuxWrite8(GMUX_PORT_INTERRUPT_ENABLE, GMUX_INTERRUPT_ENABLE);
  //
  // This code will ask GMUX to switch to Iris Pro

  GmuxWrite8(GMUX_PORT_SWITCH_DDC, 1);
  GmuxWrite8(GMUX_PORT_SWITCH_DISPLAY, 2);
  GmuxWrite8(GMUX_PORT_SWITCH_EXTERNAL, 2);

  udelay(1000);

  //
  // This code will ask GMUX to power down discrete graphics

  GmuxWrite8(GMUX_PORT_DISCRETE_POWER, 1);
  GmuxWrite8(GMUX_PORT_DISCRETE_POWER, 0);

  //
  // We're done
  
  udelay(1000);
}

/**************************************************************************
  PowerStateCallback
**************************************************************************/

VOID
  PowerStateCallback(
  __in PVOID Context,
  __in PVOID Argument1,
  __in PVOID Argument2
  )
{
  PDEV_EXT pDevExt;

  pDevExt = (PDEV_EXT) Context;

  if (Argument1 != (PVOID) PO_CB_SYSTEM_STATE_LOCK) {
    return;
  }

  if (Argument2 == (PVOID) 0) {
  
  }
  else if (Argument2 == (PVOID) 1) {
    //
    // We have reentered S0 (note that we may have never left S0)
    
    SwitchToIntelGraphicsAndPowerOffDiscreteGPU();

  }
}

/**************************************************************************
  DeregisterPowerStateChangeCallback
**************************************************************************/

NTSTATUS
  RegisterPowerStateChangeCallback(
  PDEV_EXT DevExt
  )
{
  OBJECT_ATTRIBUTES oa;
  UNICODE_STRING string;
  NTSTATUS status = STATUS_SUCCESS;

  RtlInitUnicodeString(&string, L"\\Callback\\PowerState");
  
  InitializeObjectAttributes(&oa,
    &string,
    OBJ_CASE_INSENSITIVE | OBJ_PERMANENT,
    NULL,
    NULL);

  //
  // Create a callback object, but we do not want to be the first ones
  // to create it (it should be created way before we load in NTOS
  // anyways) so we pass FALSE for the 3rd parameter.
  //
  status = ExCreateCallback(&DevExt->CbObject, &oa, TRUE, TRUE);

  if (NT_SUCCESS(status)) 
  {
    DevExt->CbRegistration = ExRegisterCallback(DevExt->CbObject,
      PowerStateCallback,
      DevExt);
    
    if (DevExt->CbRegistration == NULL) 
    {
      ObDereferenceObject(DevExt->CbObject);
      DevExt->CbObject = NULL;

      status = STATUS_UNSUCCESSFUL;
    }
  }

  return status;
}

/**************************************************************************
  DeregisterPowerStateChangeCallback
**************************************************************************/

VOID
  DeregisterPowerStateChangeCallback(
  PDEV_EXT DevExt
  )
{
  if (DevExt->CbRegistration != NULL) {
    ExUnregisterCallback(DevExt->CbRegistration);
    DevExt->CbRegistration = NULL;
  }

  if (DevExt->CbObject != NULL) {
    ObDereferenceObject(DevExt->CbObject);
    DevExt->CbObject = NULL;
  }
}

/**************************************************************************
  DriverEntry
**************************************************************************/

NTSTATUS
  DriverEntry(
  __in PDRIVER_OBJECT DriverObject,
  __in PUNICODE_STRING RegistryPath
  )
{
  NTSTATUS status = STATUS_SUCCESS;
  UNICODE_STRING UnicodeString;
  UNICODE_STRING dosDeviceName, RegName;
  UNICODE_STRING RegValueName;
  PDEVICE_OBJECT GMUXSystemDeviceObject = NULL;
  OBJECT_ATTRIBUTES RegKeyAttributes;
  HANDLE RegKey;
  UCHAR buffer[sizeof(KEY_VALUE_PARTIAL_INFORMATION) + 1024];
  ULONG resultLength;
  UCHAR intelGpuPresentAndEnabled = 0;
  PWCHAR domainStr;
  ULONG domainStrL;
  int i = 0;

  UNREFERENCED_PARAMETER(RegistryPath);

  intelGpuPresentAndEnabled = IsIntelGpuEnabled();

  //
  // Switch to Intel iGPU and power down NVIDIA graphics

  if(intelGpuPresentAndEnabled) {
    SwitchToIntelGraphicsAndPowerOffDiscreteGPU();
  }

  RtlInitUnicodeString(&UnicodeString, NT_DEVICE_NAME);
  RtlInitUnicodeString(&dosDeviceName, DOS_DEVICE_NAME);

  status = IoCreateDevice(DriverObject,
    0,
    &UnicodeString,
    FILE_DEVICE_UNKNOWN,
    FILE_DEVICE_SECURE_OPEN,
    FALSE,
    &GMUXSystemDeviceObject
    );

  if (!NT_SUCCESS(status))
    return status;

  DriverObject->DriverUnload = GMUXUnload;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = dummyFunction;
  DriverObject->MajorFunction[IRP_MJ_CREATE] = dummyFunction;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = deviceControl;

  IoCreateSymbolicLink(&dosDeviceName, &UnicodeString);

  //
  // Register for power callbacks
  
  if(intelGpuPresentAndEnabled) {
    RegisterPowerStateChangeCallback(&gDevExt);
  }

  return status;
}


NTSTATUS dummyFunction(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  UNREFERENCED_PARAMETER(DeviceObject);

  PAGED_CODE();

  Irp->IoStatus.Status = STATUS_SUCCESS;
  Irp->IoStatus.Information = 0;


  IoCompleteRequest(Irp, IO_NO_INCREMENT);

  return STATUS_SUCCESS;
}


VOID GMUXUnload(PDRIVER_OBJECT DriverObject)
{
  PDEVICE_OBJECT deviceObject = DriverObject->DeviceObject;
  UNICODE_STRING nameString;

  PAGED_CODE();

  RtlInitUnicodeString(&nameString, DOS_DEVICE_NAME);
  IoDeleteSymbolicLink(&nameString);

  if (deviceObject != NULL)
  {
    IoDeleteDevice(deviceObject);
  }
}


NTSTATUS deviceControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  NTSTATUS status = STATUS_SUCCESS;
  return status;
}
